# -*- tcl -*-
# $Id: sandbox.tcl,v 1.59 2003/08/08 16:34:39 lioka Exp $
#
# Copyright (c) 2001,2002,2003 Sergey Bolshakov <sbolshakov@altlinux.ru>
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

namespace eval sandbox {
    namespace import [namespace parent]::*
    namespace export sandbox
    variable commands [list cget destroy]
    variable counter 0
}

# dir: sandbox created under
# optional rest of args (keyed list):
# name: dirname under $dir
# size: size of sandbox loop fs, default 2GB
# fstype: TODO, ext2 hardcoded
# fsopts: args to mkfs.$fstype
proc sandbox::sandbox {dir {rest {}}} {
    namespace eval [namespace current]::[set id sb[incr [namespace current]::counter]] {
	array set options {}
    }
    upvar [namespace current]::${id}::options options

    set options(root) [file join $dir [expr {[keylget rest name name] ? $name : $id}]]
    set options(size) [expr {[keylget rest size size] ? $size : {2000}}]

    try_eval {
	# make root dir
	file mkdir $options(root)
	if {$options(size)} {
	    # non-nil size means loop-mounted filesystem
	    set loopfile [file join $options(root) $id]
	    read_file |[concat /bin/dd if=/dev/zero of=$loopfile bs=1MB count=1 \
			    seek=[incr options(size) -1] 2>@ stdout]
	    set options(loopfile) $loopfile
	    read_file |[concat /sbin/mke2fs -qF [expr {[keylget rest fsopts fsopts] ?
						       $fsopts : {}}] $options(loopfile)]
	    read_file |[concat /sbin/tune2fs -C1 -econtinue -i0 $options(loopfile)]
	    read_file |[concat /usr/bin/sudo /bin/mount -text2 -o loop \
			    $options(loopfile) $options(root)]
	    set options(ismounted) 1
	}
    } {
	set err [perror]
	destroy $id 0
	error "creating initial fs failed: $err"
    }

    # XXX creates alias in global namespace
    interp alias {} ::$id {} [namespace current]::_proc $id

    set id
}

proc sandbox::_proc {id {cmd {}} {args {}}} {
    variable commands

    if {[llength [info level 0]] == 2 } {
	error "wrong \# args: should be \"$name option ?arg arg ...?\""
    }
    if {[lsearch -exact $commands $cmd] == -1} {
	error "wrong command $cmd: should be $commands"
    } else {
	eval [list [namespace current]::$cmd $id] $args
    }
}

proc sandbox::cget {id {key {}}} {
    upvar [namespace current]::${id}::options options

    if {[string eq $key {}]} {
	array names options
    } else {
	set options($key)
    }
}

proc sandbox::destroy {id keep} {
    upvar [namespace current]::${id}::options options

    set options(keep) $keep
    destroy:lazy $id
}

proc sandbox::destroy:lazy id {
    upvar [namespace current]::${id}::options options

    set sb {}

    if {[info exists options(ismounted)] && [catch { 
	read_file |[concat /usr/bin/sudo /bin/umount $options(root)]
    } f]} {
	log dbg "\[$id\] destroy: $f"
	after 1000 [namespace code [list destroy:lazy $id]]
	return
    }

    if {$options(keep) && [info exists options(loopfile)]} {
	set sb [file dirname $options(loopfile)].[pid]
	file rename -force $options(loopfile) $sb
    }

    file delete -force $options(root)

    namespace delete [namespace current]::$id
    catch {interp alias {} ::$id {}}

    log msg "\[$id\] destroy"

    set sb
}
