# -*- tcl -*-
# $Id: sandcl.tcl,v 1.43 2003/05/13 11:06:07 lioka Exp $
#
# Copyright (c) 2001,2002,2003 Sergey Bolshakov <sbolshakov@altlinux.ru>
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

namespace eval client {
    variable usage "Usage:
sandcl \[options\] command \[command options\]
options:
-d|--repos repos - repository, \[\[user\[:password\]@\]host:\[port\]/\]repos
-h|--help        - this help"
}

###############################################################################
# public 
#

proc client::init {} {
    variable usage

    if {[string eq $::tcl_patchLevel 8.4.0]} {
	fconfigure stdin -translation auto
	fconfigure stdout -translation auto
    }

    if {!$::argc} {
	puts $usage
	exit 0
    }

    if {[info exists ::env(BTEROOT)]} {
	set bteroot $::env(BTEROOT)
    }

    while {![info exists command] && [llength $::argv]} {
	switch -exact -- [lindex $::argv 0] {
	    --repos -
	    -d {
		lvarpop ::argv
		set bteroot [lvarpop ::argv]
	    }
	    --help -
	    -h {
		puts $usage
		exit 0
	    }
	    default {
		set command "-task [lvarpop ::argv]"
	    }
	}
    }

    if {![info exists bteroot]} {
	error "-repos is not specified and BTEROOT is not set"
    }

    if {![info exists command]} {
	error "no command given\n$usage"
    }

    if {[regexp {(?x)
	^(?:([^:]+){0,1}(?::(.+)){0,1}@){0,1}
	(?:(.+):(\d+){0,1}/){0,1}
	([^/]+)$
    } $bteroot {} user pass host port repos]} {
	if {[string eq $user {}]} {
	    set user "-user [id user]"
	} else {
	    set user "-user $user"
	}
	if {[string eq $host {}]} {
	    set host localhost
	}
	if {[string eq $port {}]} {
	    set port 65432
	}
	set pass "-pass $pass"
	set repos "-repos $repos"
    } else {
	error "bad repos: $bteroot\n$usage"
    }

    set sock [socket $host $port]
    fconfigure $sock -buffering line

    switch -glob -- "$command" {
	*get* {
	    set rest $::argv
	    set dst stdout
	    set src [open |[list cpio -iumv --quiet <@ $sock |& cat]]
	}
	*add* {
	    if {[string eq [set rest [lvarpop ::argv]] {}]} {
		error "no package name given"
	    }
	    set dst $sock
	    if {[llength $::argv]} {
		pipe pr pw
		fcntl $pw CLOEXEC 1
		set src [open |[list cpio -o -H crc --quiet <@ $pr]]
		puts $pw [join $::argv \n]
		close $pw
	    } else {
		# assumed data is piped into stdin
		set src stdin
	    }
	}
	default {
	    set rest $::argv
	}
    }

    set rest "$user $pass $repos $command $rest"
    regsub -all {\s+} [string trim $rest] { } rest
    puts $sock $rest
    after 500

    if {[info exists src] && [info exists dst]} {
	fconfigure $src -translation binary
	fconfigure $dst -translation binary
	fcopy $src $dst
	fconfigure $src -translation auto
	fconfigure $dst -translation auto
    }

    fcopy $sock stdout -command [list lappend [namespace current]::done]
    if {![info exists [namespace current]::done]} {
	vwait [namespace current]::done
    }
}
