# -*- tcl -*-
# $Id: sandcvs.tcl,v 1.34 2004/09/13 14:05:03 me Exp $
#
# Copyright (c) 2001,2002,2003 Sergey Bolshakov <sbolshakov@altlinux.ru>
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

namespace eval cvs {
    namespace import [namespace parent]::*
    namespace export cvs
    
    upvar [namespace parent]::layout layout
    variable commands [list auth user cget hook destroy \
			   names versions releases module spec \
			   ismodule branchname release diff]
    variable hooks [list dump commit tag]
    variable counter 0
}

###############################################################################
# public
#
proc cvs::cvs opts {
    variable layout

    namespace eval [namespace current]::[set id cvs[incr [namespace current]::counter]] {
	array set options {}
	set packages {}
    }

    upvar [namespace current]::${id}::options options

    set options(name) [lvarpop opts]
    set cvsroot [lvarpop opts]
    set options(cvs) [concat /usr/bin/cvs -d $cvsroot]
    # local CVSROOT stays unmodified
    regexp {^[^/]+(.+)$} $cvsroot {} cvsroot
    # cutout all doubled & trailing slashes
    set options(cvsroot) [file normalize $cvsroot]

    if {[scan [lvarpop opts] {%[^, ]%*[, ]%[^, ]} options(module) options(branch)] == 2} {
	set options(branchcmd) [list -r $options(branch)]
    } else {
	set options(branch) head
	set options(branchcmd) {}
    }

    set options(tmp) [file join $layout(tmpdir) [file tail $::argv0] $id.[pid]]

    if 0 {
	log dbg "--- $options(name) ---"
	foreach e [lsort [array names options]] {
	    log dbg "$e: $options($e)"
	}
    }

    # inits packages struct
    init $id

    # XXX creates alias in global namespace
    interp alias {} ::$id {} [namespace current]::_proc $id
    set id
}

proc cvs::_proc {id {cmd {}} {args {}}} {
    variable commands

    if {[llength [info level 0]] == 2 } {
	error "wrong \# args: should be \"$id option ?arg arg ...?\""
    }
    if {[lsearch -exact $commands $cmd] == -1} {
	error "wrong command $cmd: should be $commands"
    } else {
	eval [list [namespace current]::$cmd $id] $args 
    }
}

###############################################################################
# commands
###############################################################################

proc cvs::destroy id {
    namespace delete [namespace current]::$id
    interp alias {} ::$id {}
}

proc cvs::auth {id user pass} {
    upvar [namespace current]::${id}::options options

    set sc [scancontext create]
    set fd [open |[concat $options(cvs) -Q co -p CVSROOT/passwd.$options(name) CVSROOT/passwd]]
    scanmatch $sc "^$user:(\[^:\]*)(?::.+){0,1}$" {
	set hash [crypt $pass [string range $matchInfo(submatch0) 0 1]]
	if {[string eq $hash $matchInfo(submatch0)]} {
	    set auth_passed 1
	}
	break
    }
    scanfile $sc $fd
    scancontext delete $sc
    catch {close $fd}

    if {![info exists auth_passed]} {
	error "authorization failed for $user"
    }
}

proc cvs::user {id user} {
    upvar [namespace current]::${id}::options options

    set sc [scancontext create]
    set fd [open |[concat $options(cvs) -Q co -p CVSROOT/users.$options(name) CVSROOT/users]]
    scanmatch $sc "$user:\"(.+)\"" {
	set uinfo $matchInfo(submatch0)
	break
    }

    scanfile $sc $fd
    scancontext delete $sc
    catch {close $fd}
	
    if {![info exists uinfo]} {
	error "invalid user `$user'"
    }

    set uinfo
}

proc cvs::names {id {relid {}}} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    set names [keylkeys packages]
    if {[set ndx [lsearch -exact $names *]] != -1} {
	lvarpop names $ndx
    }

    if {$relid ne {}} {
	if {![lcontain [keylkeys options(releases)] $relid]} {
	    error "no such release id $relid"
	}
	set nn {}
	foreach n $names {
	    set releases [rels $id $n]
	    if {[lcontain [keylkeys releases] $relid]} {
		lappend nn $n
	    }
	}
	set names $nn
    }

    split [string map {^ .} [join $names]]
}

proc cvs::versions {id name {relid {}}} {
    upvar [namespace current]::${id}::packages packages

    if {$relid eq {}} {
	tags $id [string map {. ^} $name]
    } else {
	set releases [rels $id [string map {. ^} $name]]
	expr {[keylget releases $relid ver] && $ver ne {} ? $ver : {undefined}}
    }
}

proc cvs::releases id {
    upvar [namespace current]::${id}::options options
    
    lsort -index 0 $options(releases)
}

proc cvs::cget {id option mixin} {
    upvar [namespace current]::${id}::options options

    read_file |[concat $options(cvs) -Q co $options(branchcmd) -p \
		    [
		     switch -exact -- $option {
			 macros {
			     list \
				 [file join $options(module) macros] \
				 [file join $options(module) macros.$mixin]
			 }
			 release {
			     list \
				 [file join $options(module) release] \
				 [file join $options(module) release.$mixin]
			 }
			 installer -
			 updater {
			     list \
				 [file join $options(module) release] \
				 [file join $options(module) release.$mixin] \
				 [file join $options(module) $option.$mixin] \
				 [file join $options(module) $option]
			 }
			 default {
			     error "unknown option $option"
			 }
		     }
		    ]
	       ]
}

proc cvs::spec {id name {version current}} {
    upvar [namespace current]::${id}::options options

    read_file |[concat $options(cvs) -Q co -p \
		    [rev $id $name $version] \
		    [path_byname $id $name]]
}

proc cvs::module {id path name {version current}} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    set here [pwd]
    set dir [file dirname $path]
    set ddir [file tail $path]
    try_eval {
	file mkdir $dir
	cd $dir
	read_file |[concat $options(cvs) -Q export -d $ddir [rev $id $name $version] \
			[expr {[keylget packages [string map {. ^} $name].module module] ?
			       $module : $name}]]
    } {
	catch {file delete -force -- $path}
	error "couldn't export $name to $path: [perror]"
    } {
	cd $here
    }
}

proc cvs::branchname {id name} {
    upvar [namespace current]::${id}::packages packages

    set branchname [branch_byname $id $name]
    expr {$branchname eq {HEAD} ? {} : $branchname}
}

proc cvs::ismodule {id name} {
    upvar [namespace current]::${id}::packages packages
    
    expr {$name ne {*}} && \
	[keylget packages [string map {. ^} $name].module {}]
}

proc cvs::diff {id fd rest} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    if {[llength $rest] != 2} {
	error "cvs::diff: expected 2 relid, got $rest"
    }

    foreach {ina inboth inb} [intersect3 [names $id [lindex $rest 0]] [names $id [lindex $rest 1]]] break
    if {[lempty $inboth]} {
	error "cvs::diff: totally differs, can't proceed"
    }

    foreach name $inboth {
	lappend cvspath [expr {[ismodule $id $name] ?
			       [keylget packages $name.module] :
			       [file join [keylget packages *.module] [string map {^ .} $name]]}
			]
    }

    # FIXME
    catch {
	popen [concat $options(cvs) -Q rdiff -u \
		   -r $options(branch)-[lindex $rest 0]-release \
		   -r $options(branch)-[lindex $rest 1]-release \
		   [lsort -uniq $cvspath]] $fd
    }
}

proc cvs::hook {id hook callback path rest} {
    variable hooks

    if {![lcontain $hooks $hook]} {
	    error "unknown hook cvs:$hook"
    }
	
    $hook $id $callback $path $rest
}

###############################################################################
# private
###############################################################################
# inits options(releases)
# inits packages keyed list, each element is a:
# pkgname {
#         branch <branch name>
#         module <module path> # optional
#         tags { # actually versions
#              <tag name 0>  <revision 0>
#              <tag name 1>  <revision 1>
#              ...
#         }
#         rels { # release tags
#              <tag name 0>  <revision 0>
#              <tag name 1>  <revision 1>
#         }
# }
proc cvs::init id {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages
    
    set here [pwd]
    set options(releases) {}
    set relpath [file join $options(module) release]
    try_eval {
	file mkdir $options(tmp)
	cd $options(tmp)
	# FIXME Attic complains to stderr
	read_file |[concat $options(cvs) -Q co $relpath]
	try_eval {
	    set sc [scancontext create]
	    set fd [open |[concat $options(cvs) -Q log -S $relpath]]
	    scanmatch $sc {(?i)symbolic names:} {
		scanmatch $sc [subst -noc -nob {^\s+$options(branch)-([^-]+)-release:\s+(\S+)}] {
		    lappend releases $matchInfo(submatch0) $matchInfo(submatch1)
		}
	    }
	    scanmatch $sc {(?i)total revisions:} {
		if {![info exists releases]} return
	    }
	    scanmatch $sc [subst -noc -nob {^revision\s(.+)$}] {
		set rev $matchInfo(submatch0)
		scanmatch $sc {date:\s+([^;]+)} {
		    set date $matchInfo(submatch0)
		    while {[set ndx [lsearch -exact $releases $rev]] != -1} {
			keylset options(releases) [lvarpop releases [incr ndx -1]] [string map {/ -} $date]
			lvarpop releases $ndx
		    }
		    continue
		}
	    }
	    scanfile $sc $fd
	} {
	    error "couldn't stat `$relpath': [perror]"
	} {
	    scancontext delete $sc
	    catch {close $fd}
	}
    } {
	error [perror]
    } {
	file delete -force $options(tmp)
	cd $here
    }
    
    set modpath [file join $options(module) modules]
    try_eval {
	set fd [open |[concat $options(cvs) -Q co $options(branchcmd) -p $modpath]]
	parse $fd modules
	if {[keylget modules $options(name) modules]} {
	    set packages {}
	    foreach n [keylkeys modules] {
		foreach {path branch} [scan [keylget modules $n] {%[^, ]%*[, ]%[^, ]}] break
		keylset packages \
		    ${n}.module $path \
		    ${n}.branch [expr {[lempty $branch] ? {HEAD} : $branch}]
	    }
	} else {
	    error "no section \[$options(name)\] in $modpath"
	}
    } {
	error [perror]
    } {
	catch {close $fd}
    }

    set packagepath [keylget packages *.module]
    set packagebranch [keylget packages *.branch]
    try_eval {
	set sc [scancontext create]
	# FIXME tooooo slooooooow
	set fd [open |[concat $options(cvs) co -r $packagebranch -p $packagepath 2>@ stdout]]
	# skip things like .cvsignore
	scanmatch $sc "^Checking out $packagepath/(\[^\.\].+)$" {
	    keylset packages [string map {. ^} $matchInfo(submatch0)] {}
	}
	scanfile $sc $fd
    } {
	error [perror]
    } {
	scancontext delete $sc
	catch {close $fd}
    }
    
    if 0 {
	log dbg "-------- $options(name) --------"
	log dbg "releases: $options(releases)"
	foreach n [lsort [keylkeys packages]] {
	    log dbg "$n: [keylget packages $n]"
	}
    }
}

# returns requested revision (by tag)
proc cvs::rev {id name version} {
    upvar [namespace current]::${id}::options options

    log msg "\[$id\] getting `$version' version of $name"
    switch -exact -- $version {
	current {
	    branch_to_cmd $id $name
	}
	tagged {
	    set alltags [tags $id $name]
	    if {![lempty $alltags]} {
		list -r [branch_to_idx $id $name]-[lindex [lindex $alltags end] 0]
	    } else {
		branch_to_cmd $id $name
	    }
	}
	default {
	    set alltags [tags $id $name]
	    set tag [string map {. _} $version]
	    if {[keylget alltags $tag rev]} {
		if {[ismodule $id $name]} {
		    list -r [branch_to_idx $id $name]-$tag
		} else {
		    # spec-only package
		    # numeric revision is enough i hope
		    list -r $rev
		}
	    } else {
		error "version `$version' of $name not exists"
	    }
	}
    }
}

###############################################################################
# tags, branches, releases
###############################################################################
proc cvs::tags {id name} {
    upvar [namespace current]::${id}::packages packages

    if {[keylget packages [string map {. ^} $name].tags tags]} {
	return $tags
    }

    scantags $id $name
    keylget packages [string map {. ^} $name].tags
}

proc cvs::rels {id name} {
    upvar [namespace current]::${id}::packages packages

    if {[keylget packages [string map {. ^} $name].rels rels]} {
	return $rels
    }

    scantags $id $name
    keylget packages [string map {. ^} $name].rels
}

###############################################################################
# scans version & release tags
# constructs tag klist:
# branch1 {{tag revision} {tag revision} ...}
# branch2 {{tag revision} ...}
# ...
# head {{tag revision} ...}
# tag-rev lists in each branch or head soted by tags,
# most recent last
# and appends both to packages klist
# as tags & rels
proc cvs::scantags {id name} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    # pass one: collect raw tag/branch info
    set revs {}
    set branches {}
    set here [pwd]
    try_eval {
	file mkdir $options(tmp)
	cd $options(tmp)
	# FIXME Attic complains to stderr
	read_file |[concat $options(cvs) -Q co [path_byname $id $name]]
	try_eval {
	    set sc [scancontext create]
	    set fd [open |[concat $options(cvs) -Q stat -v [path_byname $id $name]]]
	    scanmatch $sc {(?i)existing tags:} {
		scanmatch $sc {\s+(\S+)\s+\(branch:\s+(.+)\)} {
		    lappend branches [list $matchInfo(submatch0) $matchInfo(submatch1)]
		}
		scanmatch $sc {\s+(\S+)\s+\(revision:\s+(.+)\)} {
		    lappend revs $matchInfo(submatch0),$matchInfo(submatch1)
		}
	    }
	    scanfile $sc $fd
	} {
	    error "couldn't checkout `$name': [perror]"
	} {
	    scancontext delete $sc
	    catch {close $fd}
	}
    } {
	error [perror]
    } {
	file delete -force $options(tmp)
	cd $here
    }


    # pass two: separate version tags and release tags
    # with sorting by branches
    set version_tags {}
    set release_tags {}

    # search for tags in branches, 'long' branches first
    foreach branch [lsort -dec -index end $branches] {
	set v {}
	set r {}
	set bt [lindex $branch 0]
	set br [lindex $branch 1]
	while {[set ndx [lsearch $revs *,$br.*]] != -1} {
	    foreach {it ir} [split [lindex $revs $ndx] ,] break
	    if {[regsub [subst -noc -nob {^$bt-(\d-[^-]+-[^-]+)$}] $it {\1} it]} {
		# package version tag
		lappend v [list $it $ir]
	    }
	    if {[regsub [subst -noc -nob {^$options(branch)-([^-]+)-release$}] $it {\1} it]} {
		# release tag
		lappend r [list $it $ir]
	    }
	    lvarpop revs $ndx
	}
	lappend version_tags [list $bt [lsort -dictionary -uniq -index 0 $v]]
	lappend release_tags [list $bt [lsort -dictionary -uniq -index 0 $r]]
    }

    # assuming remaining tags are in HEAD
    set v {}
    set r {}
    foreach rv $revs {
	foreach {it ir} [split $rv ,] break
	if {[regsub {head-(\d-[^-]+-[^-]+)$} $it {\1} it]} {
	    lappend v [list $it $ir]
	}
	if {[regsub [subst -noc -nob {^$options(branch)-([^-]+)-release$}] $it {\1} it]} {
	    # release tag
	    lappend r [list $it $ir]
	}
    }

    lappend version_tags [list head [lsort -dictionary -uniq -index 0 $v]]
    lappend release_tags [list head [lsort -dictionary -uniq -index 0 $r]]

    # pass three :
    # omit not ours according to `modules' settings
    # replaces revisions in release list to version tags
    # links both to packages
    # FIXME pass the rest to other cvs instances ?
    set v {}
    set r {}

    keylget version_tags [branch_to_idx $id $name] v
    keylget release_tags [branch_to_idx $id $name] r

    foreach rel [keylkeys r] {
	set notfound 1
	foreach ver [keylkeys v] {
	    if {[keylget r $rel] eq [keylget v $ver]} {
		keylset r $rel $ver
		set notfound 0
		break
	    }
	}
	if {$notfound} {
	    keylset r $rel {}
	}
    }

    keylset packages [string map {. ^} $name].tags $v
    keylset packages [string map {. ^} $name].rels $r
}

###############################################################################
# dancing with release marks
###############################################################################
# strange things happens here:
# in: keyed list of all exported packages
# filters out all unknown by this repo package names
# then marks appropriate versions(by tags) of names
# and the whole `release' module with given mark
proc cvs::release {id release names} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    set release [string map {. _} $release]
    set mark $options(branch)-$release-release

    foreach name [keylkeys names] {
	# skip foreign packages
	if {![lcontain [keylkeys packages] $name]} continue
	set version \
	    [string map {. _} \
		 [expr {[set serial [keylget names $name.serial]] ne {(none)} ?
			[set serial] : {0}}
		 ]-[
		    keylget names $name.version
		   ]-[
		      keylget names $name.release
		     ]
	    ]
	set alltags [tags $id [string map {^ .} $name]]
	if {[lsearch -exact [keylkeys alltags] $version] == -1} {
	    set tag [branch_byname $id $name]
	    set version {}
	} else {
	    set tag [version_to_tag $id $name $version]
	}

	lappend rest $name $version $release $tag $mark \
	    [expr {[ismodule $id $name] ?
		   [keylget packages $name.module] :
		   [file join [keylget packages *.module] [string map {^ .} $name]]}
	    ]
    }

    if {![info exists rest]} return

    # first mark cvsmodule
    popen [concat $options(cvs) -Q rtag $options(branchcmd) $mark $options(module)]
    keylset options(releases) $release [clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S} -gmt 1]
    after idle [list [namespace current]::release:post $id $rest]
}

proc cvs::release:post {id rest} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    foreach v {name version release tag mark module} {
	set $v [lvarpop rest]
    }

    # XXX deadlock with read_file
    if {[catch {popen [concat $options(cvs) -Q rtag -r $tag $mark $module]} f]} {
	# XXX push it back to list ? after N times ?
	log msg "\[$id\] setting tag `$mark' to `$tag' of $module failed: $f"
    } else {
	keylset packages $name.rels \
	    [lsort -dictionary -index 0 \
		 [concat [rels $id $name] [list [list $release $version]]]]
	log msg "\[$id\] setting tag `$mark' to `$tag' of `$module'"
    }
    # poor man's tail recursion :/
    if {![lempty $rest]} {
	after idle [list [namespace current]::release:post $id $rest]
    }
}

###############################################################################
# CVS hooks
###############################################################################
# hooks from CVSROOT/commitinfo
###############################################################################
proc cvs::commit {id callback path rest} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    # check for commit triggered in one of our modules
    if {[catch {checkrep $id $path [lvarpop rest]} cvspath]} {
	log msg "\[$id\] declined: $cvspath"
	# not our module, continue search
	return 0
    }

    # TODO add hook for release{,.xxx}
    if {$cvspath eq $options(module)} {
	if {[lcontain $rest modules]} {
	    log msg "\[$id\] rebuilding administrative file database"
	    after 5000 [list [namespace current]::init $id]
	}
	return 1
    }

    if {$cvspath ne [keylget packages *.module] &&
	[set name [name_bypath $id $cvspath]] eq {}} {
	# found, but not in module topdir, stop search
	return 1
    }

    if {[info exists name]} {
	if {![lcontain $rest $name.spec]} {
	    # module's topdir, w/o spec file
	    return 1
	} elseif {[llength $rest] != 1} {
	    error "extra files not allowed when commiting spec: $rest"
	}
    }

    # ignore dotted files i.e. .cvsignore
    while {[set idx [lsearch -glob $rest .*]] != -1} {
	lvarpop rest $idx
    }

    set taglist {}
    # grab current revisions
    for_file l [file join $path CVS Entries] {
	if {[scan $l {/%[^/]/%[^/]} f r] != 2} continue
	if {![lcontain $rest $f]} continue
	if {[string match -* $r]} {
	    log msg "\[$id\] removing entry: $f"
	    keyldel packages [string map {. ^} [expr {[info exists name] ? $name : $f}]]
	    continue
	}
	# check spec for consistency and return tag info if needed
	set taglist [concat $taglist [eval $callback [file join $path $f] $r]]
	
	if {$r eq {0}} {
	    log msg "\[$id\] adding entry: $f"
	    keylset packages [string map {. ^} [expr {[info exists name] ? $name : $f}]] {}
	}
    }

    if {![lempty $taglist]} {
	after 1000 [list [namespace current]::commit:post $id $taglist]
    }

    return 1
}

proc cvs::commit:post {id tags} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    set name [lvarpop tags]
    set version [lvarpop tags]
    set revision [lvarpop tags]
    set tag [version_to_tag $id $name $version]
    if {[ismodule $id $name]} {
	set rev [branch_to_cmd $id $name]
	set module [keylget packages $name.module]
    } else {
	set rev [list -r $revision]
	set module [file join [keylget packages *.module] $name]
    }
    # XXX deadlock with read_file
    if {![catch {popen [concat $options(cvs) -Q rtag $rev $tag $module]} f]} {
	keylset packages [string map {. ^} $name].tags \
	    [lsort -dictionary -index 0 \
		 [concat [tags $id $name] \
		      [list [list [string map {. _ } $version] $revision]]]]
    } else {
	# XXX push it back to list ? after N times ?
	log msg "\[$id\] setting tag $tag to revision $revision of $module failed: $f"
    }

    # poor man's tail recursion :/
    if {![lempty $tags]} {
	after idle [list [namespace current]::spawntag $id $tags]
    }
}

###############################################################################
# hooks from CVSROOT/tagingo
# TODO hardening tag policy
###############################################################################
proc cvs::tag {id callback path rest} {
    # check for tag triggered in one of our modules
    if {[catch {checkrep $id $path [lvarpop rest]} cvspath]} {
	log msg "\[$id\] declined: $cvspath"
	# not our module, continue search
	return 0
    }

    log msg "\[$id\] tag: $rest"
    return 1
}

proc cvs::checkrep {id path cvspath} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    set cvspath [file normalize $cvspath]
    if {![string match $options(cvsroot)/* $cvspath]} {
	error "$cvspath: not under cvsroot $options(cvsroot)"
    }
    
    set cvspath [string range $cvspath [expr [string length $options(cvsroot)] + 1] end]

    if {$cvspath ne $options(module)} {
	set modpath $cvspath
	while {$modpath ne {.} && [set name [name_bypath $id $modpath]] eq {}} {
	    set modpath [file dirname $modpath]
	}
	
	if {$modpath eq {.}} {
	    error "$cvspath: not in repository $options(name)"
	}
    }

    if {[file exists [set branch [file join $path CVS Tag]]] &&
	[scan [read_file $branch] {T%s} branch]} {
        if {$cvspath eq $options(module)} {
	    if {$options(branch) ne $branch} {
		error "branch $branch: not in repository $options(name)"
	    }
	} else {
	    if {$branch ne [keylget packages $name.branch]} {
		error "branch $branch: not in repository $options(name)"
	    }
	}
    }

    set cvspath
}

proc cvs::dump {id callback path rest} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    foreach name [keylkeys packages] {
	eval $callback [list [list $name [keylget packages $name]]]
    }

    return 1
}

###############################################################################
# helpers
###############################################################################
proc cvs::name_bypath {id path} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    if {$path eq [keylget packages *.module]} {
	return *
    }

    foreach n [keylkeys packages] {
	if {[keylget packages $n.module module] && $path eq $module} {
	    return [string map {^ .} $n]
	}
    }
}

proc cvs::path_byname {id name} {
    upvar [namespace current]::${id}::options options
    upvar [namespace current]::${id}::packages packages

    if {[keylget packages [string map {. ^} $name].module module]} {
	file join $module [string map {^ .} $name].spec
    } else {
	file join [keylget packages *.module] [string map {^ .} $name]
    }
}

proc cvs::branch_byname {id name} {
    upvar [namespace current]::${id}::packages packages

    keylget packages [expr {$name ne {*} && [ismodule $id $name] ? $name : {*}}].branch
}

proc cvs::branch_to_cmd {id name} {
    list -r [branch_byname $id $name]
}

proc cvs::branch_to_idx {id name} {
    # FIXME i hate shifts
    string tolower [branch_byname $id $name]
}

proc cvs::version_to_tag {id name version} {
    subst [branch_to_idx $id $name]-[string map {. _} $version]
}
