# -*- tcl -*-
# $Id: sandtask.tcl,v 1.107 2003/08/13 18:09:59 lioka Exp $
#
# Copyright (c) 2001,2002,2003 Sergey Bolshakov <sbolshakov@altlinux.ru>
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

namespace eval task {
    namespace import [namespace parent]::*
    namespace import [namespace parent]::rep::*
    namespace export hook task
    variable tasks [list addsources getsources build check \
			getpocket endpocket killpocket \
			exportlist exportiso exportupdate \
			query querynames queryver querybroken \
			statsreleases statsnames statsversions statsdiff statslog \
			help]
    variable defaults \
	[list \
	     {reuse {try to reuse sandbox on build}} \
	     {nonstop {don't stop on build errors}} \
	     {keepfailed {keep failed sanboxes}} \
	     {rebuild {build from existing source package}} \
	     {pocket.arg {} {pocket repository, used with `build'}} \
	     {buildreq.arg {} {additional build requirements, used with `build'}} \
	     {kernel.arg {} {assumed kernel version, used with `build'}} \
	     {mixin.arg {} {mixin id to apply}} \
	     {regexp {treat rest of args as regexp patterns, default glob}} \
	     {pkgver.arg {} {package version-release}} \
	     {relid.arg {} {stats since or between release id(s), see `stats*'}} \
	     {isoprefix.arg {} {optional prefix for cd image name}} \
	     {isosuffix.arg {} {optional suffix for cd image name}} \
	     {verbose.arg {basebox,buildbox,checkinstall,prep,build,iso} {be verbose on specified stage(s)}} \
	     {quiet {be really quiet}} \
	     {repos.arg.secret {} {repository to use}} \
	     {user.arg.secret {} username} \
	     {pass.arg.secret {} password} \
	     {handler.arg.secret {} {output channel}} \
	     {path.arg.secret {} {client data under path}} \
	     {task.arg.secret {} {task to execute, see -task help for full list}}]
    variable hdefaults \
	[list \
	     {repos.arg {} {repository to use}} \
	     {nonstop {don't stop on hook errors}} \
	     {handler.arg.secret {} {output channel}} \
	     {path.arg.secret {} {client data under path}} \
	     {hook.arg.secret {} {hook to execute}}]
    variable counter 0
}

###############################################################################
# public
#
proc task::hook opts {
    variable hdefaults

    try_eval {
	array set options [cmdline::getoptions opts $hdefaults]
	set options(rest) $opts
    } {
	if {[llength $opts]} {
	    error "unknown option `[lindex $opts 0]'"
	} else {
	    # we don't want stack trace here anyway
	    error $errorInfo
	}
    }

    if {$options(hook) eq {}} {
	error "badly formed hook"
    }

    set catched 0
    foreach options(repos) [expr {$options(repos) ne {} ? $options(repos) : [rep options]}] {
        # loop over all reps until break
        set rep [rep options]
        try_eval {
            if {[set catched [$rep hook $options(rest)]]} break
        } {
            error [perror]
        } {
            $rep destroy
        }
    }
    
    if {!$catched} {
	error "$options(hook) wasn't catched by any repos"
    }
}

proc task::task opts {
    variable defaults
    variable tasks

    namespace eval [namespace current]::[set id task[incr [namespace current]::counter]] {
	array set options {}
    }
    upvar [namespace current]::${id}::options options

    try_eval {
	array set options [cmdline::getoptions opts $defaults]
	set options(rest) $opts
    } {
	namespace delete [namespace current]::$id
	if {[llength $opts]} {
	    error "unknown option `[lindex $opts 0]'"
	} else {
	    # we don't want stack trace here anyway
	    error $errorResult
	}
    }

    foreach parm {repos user pass} {
	if {$options($parm) eq {}} {
	    namespace delete [namespace current]::$id
	    error "no `$parm' given"
	}
    }

    set task $options(task)

    # check task
    if {[lsearch -exact $tasks $task] == -1} {
	namespace delete [namespace current]::$id
	error "task `$task' not known, can be `$tasks'"
    }

    if {[string eq $task help]} {
	puts $options(handler) "known commands:\n[join $tasks \n]"
	namespace delete [namespace current]::$id
	return
    }

    set options(buildreq) [split [string trim $options(buildreq)] ,]
    set options(relid) [split [string trim $options(relid)] ,]

    if {$options(quiet)} {
	set options(verbose) {}
    } else {
	set options(verbose) [split [string trim $options(verbose)] ,]
    }

    if {$options(regexp)} {
	set options(match) -regexp
    } else {
	set options(match) -glob
    }

    unset options(regexp) options(quiet) options(task)

    try_eval {
	set rep [rep options]
	eval $task $id $rep
    } {
	error [perror]
    } {
	catch {$rep destroy}
	namespace delete [namespace current]::$id
    }
}

###############################################################################
# private
#
###############################################################################
###############################################################################
# build
#
proc task::build {id rep} {
    upvar [namespace current]::${id}::options options

    set do_reuse 0
    
    foreach package [expr {$options(rebuild) ?
			   $options(rest) : [$rep query names $options(rest)]}] {
	puts $options(handler) "Building $package ..."
	try_eval {
	    switch -glob -- $options(reuse),$do_reuse {
		0,* {
		    set repx [rep options]
		    try_eval {
			$repx box
		    } {
			error "can't create base system: [perror]"
		    }
		}
		1,0 {
		    set repx $rep
		    try_eval {
			$repx box
		    } {
			error "can't create base system: [perror]"
		    }
		    set do_reuse 1
		}
		1,1 {
		    set repx $rep
		}
		default {
		    error "reuse can't happen"
		}
	    }
	    
	    try_eval {
		$repx prep $package
	    } {
		error "can't prepare `$package': [perror]"
	    }
	    
	    try_eval {
		$repx build $package
	    } {
		error "can't build `$package': [perror]"
	    }
	} {
	    if {$options(nonstop)} {
		catch {puts $options(handler) [perror]}
	    } else {
		error [perror]
	    }
	} {
	    if {$options(reuse)} {
		# noop
	    } else {
		catch {$repx destroy}
	    }
	}
    }
}

###############################################################################
# pocket manipulations
#
proc task::getpocket {id rep} {
    upvar [namespace current]::${id}::options options

    if {[string eq $options(pocket) {}]} {
	if {[lempty $options(rest)]} {
	    error "empty pocket"
	} else { 
	    set options(pocket) [lvarpop options(rest)]
	}
    }

    $rep getpocket $options(pocket) $options(rest)
}

proc task::endpocket {id rep} {
    upvar [namespace current]::${id}::options options

    if {[string eq $options(pocket) {}]} {
	if {[lempty $options(rest)]} {
	    error "empty pocket"
	} else { 
	    set options(pocket) [lvarpop options(rest)]
	}
    }

    $rep freepocket $options(pocket)
}

proc task::killpocket {id rep} {
}

###############################################################################
# sources
#

# noop
proc task::add {id rep} {
    upvar [namespace current]::${id}::options options

    if {[lempty $options(rest)]} {
	error {usage: add <package name>}
    }

    $rep add [lindex $options(rest) 0]
}

proc task::addsources {id rep} {
    upvar [namespace current]::${id}::options options

    if {[lempty $options(rest)]} {
	error "usage: addsources <package name> ?source?"
    }

    $rep addsources [lindex $options(rest) 0]
}

proc task::getsources {id rep} {
    upvar [namespace current]::${id}::options options

    if {[lempty $options(rest)]} {
	error "usage: getsources <package name> ?mask?"
    }

    $rep getsources $options(rest)
}

###############################################################################
# checks
#
proc task::check {id rep} {
    upvar [namespace current]::${id}::options options

    $rep custombox $options(rest)
}

###############################################################################
# export variants
#
proc task::exportlist {id rep} {
    upvar [namespace current]::${id}::options options

    puts $options(handler) [join [$rep export $options(rest)] \n]
}

proc task::exportiso {id rep} {
    upvar [namespace current]::${id}::options options

    $rep iso installer $options(rest)
}

proc task::exportupdate {id rep} {
    upvar [namespace current]::${id}::options options

    $rep iso updater $options(rest)
}

###############################################################################
# query
#
proc task::query {id rep} {
    upvar [namespace current]::${id}::options options

    $rep query stats $options(rest)
}

proc task::querybroken {id rep} {
    upvar [namespace current]::${id}::options options

    $rep query unmet
}

proc task::querynames {id rep} {
    upvar [namespace current]::${id}::options options

    puts $options(handler) [join [$rep query names $options(rest)] \n]
}

proc task::queryver {id rep} {
    upvar [namespace current]::${id}::options options
    
    foreach p [$rep query names $options(rest)] {
	puts $options(handler) "$p: [join [$rep query versions $p]]"
	flush $options(handler)
    }
}

###############################################################################
# statistics:
# statsbuilds statsnames statsdiff statslog \


proc task::statsreleases {id rep} {
    upvar [namespace current]::${id}::options options

    set ret {}
    foreach k [$rep stats releases] {
	append ret [join $k] \n
    }

    puts $options(handler) $ret
}

proc task::statsnames {id rep} {
    upvar [namespace current]::${id}::options options

    if {[lempty $options(relid)]} {
	error "usage: statsnames -relid <release id>"
    }

    puts $options(handler) [join [$rep stats names] \n]
}

proc task::statsversions {id rep} {
    upvar [namespace current]::${id}::options options

    if {[lempty $options(relid)]} {
	error "usage: statsversions -relid <release id>"
    }

    foreach p [$rep stats names] {
	puts $options(handler) "$p: [join [$rep stats versions $p]]"
	flush $options(handler)
    }
}

proc task::statsdiff {id rep} {
    upvar [namespace current]::${id}::options options

    if {[llength $options(relid)] < 2} {
	error "usage: statsdiff -relid <id_1,id_2>"
    }

    $rep stats diff
    return
}

proc task::statslog {id rep} {
    upvar [namespace current]::${id}::options options

    $rep stats log $options(rest)
}
